#
# Copyright (C) 2022 Soleta Networks <info@soleta.eu>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

import os
import logging

import fdisk


def get_disks():
    """
    Walks /sys/block/ and returns files starting with 'sd',
    'nvme' or 'vd'
    """
    return sorted([ dev for dev in os.listdir('/sys/block/')
                    if dev.startswith('sd')
                    or dev.startswith('nvme')
                    or dev.startswith('vd')])


def get_partition_device(disknum, partnum):
    """
    Returns the device path, given a disk and partition number
    """
    disk_index = disknum - 1
    if disk_index < 0 or disk_index >= len(get_disks()):
        raise ValueError(f'Invalid disk number {disknum}, {len(get_disks())} disks available.')

    disk = get_disks()[disk_index]
    cxt = fdisk.Context(f'/dev/{disk}')

    for pa in cxt.partitions:
        if pa.partno == partnum - 1:
            return cxt.partition_to_string(pa, fdisk.FDISK_FIELD_DEVICE)

    raise ValueError(f'No such partition with disk index {disknum} and partition index {partnum}')


def get_efi_partition(disknum, enforce_gpt):
    """
    Look for an EFI System Partition at the n-th disk. If disknum is invalid an
    exception is thrown.
    If enforce_gpt is set to True the ESP will be ignored in a MBR partition
    scheme.

    Returns tuple with:
        - Device name containing the ESP
        - /dev/{device} string
        - Partition number (starting at 1)
    """
    disk_index = disknum - 1
    if disk_index < 0 or disk_index >= len(get_disks()):
        raise ValueError(f'Invalid disk number {disknum} when trying to find ESP, {len(get_disks())} disks available.')

    disk = get_disks()[disk_index]
    cxt = fdisk.Context(f'/dev/{disk}')

    if enforce_gpt and cxt.label == fdisk.FDISK_DISKLABEL_DOS:
        raise RuntimeError(f'Disk has DOS partition scheme, cannot find ESP at /dev/{disk}')

    for pa in cxt.partitions:
        logging.info(f'Checking partition "{pa.type.name}"...')
        if pa.type.name == 'EFI System':
            return cxt.partition_to_string(pa, fdisk.FDISK_FIELD_DEVICE), f'/dev/{disk}', pa.partno + 1
    raise RuntimeError(f'Cannot find "EFI System" partition at /dev/{disk}')
